class ForecasterException(Exception):
    pass


class EstimatorException(Exception):
    pass


class ModelException(Exception):
    pass


class IllegalArgument(Exception):
    pass


class DataInputException(Exception):
    pass


class PredictionException(Exception):
    pass


class BacktestException(Exception):
    pass


class AbstractMethodException(Exception):
    pass


class PlotException(Exception):
    pass
